/* Copyright 2008 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: MainState.cpp,v 1.9 2008-04-21 00:00:55 tonic Exp $
 * $Revision: 1.9 $
 */

#include <assert.h>
#include <algorithm>

#include "main.h"
#include "MainState.h"


#define TIME_SCALE 0.5f

#define PLAYER_RADIUS 0.9f
#define BULLET_RADIUS 0.2f

#define ENEMY_DEATH_DURATION 15


static f32 sEnemyRadii[ET_COUNT] =
{
    0.63f, // 0 tetrahedron
    1.18f, // 1 cube
    1.22f, // 2 dodecahedron, descending
    1.38f, // 3 octahedron
    1.6f,  // 4 icosahedron, descending
    1.71f  // 5 octotoad
};

static s32 sEnemyAges[ET_COUNT] =
{
    330,
    900,
    1200,
    1200,
    1800,
    3000
};

enum SAMPLE_ID
{
    SAMPLE_EXPLOSION_FIRST = 0,
    SAMPLE_EXPLOSION_COUNT = 6,
    SAMPLE_EXPLOSION_LAST = SAMPLE_EXPLOSION_FIRST + SAMPLE_EXPLOSION_COUNT - 1,
    SAMPLE_SHOOT_FIRST,
    SAMPLE_SHOOT_COUNT = 4,
    SAMPLE_SHOOT_LAST = SAMPLE_SHOOT_FIRST + SAMPLE_SHOOT_COUNT - 1,
    SAMPLE_HIT,
    SAMPLE_END,
    SAMPLE_E2_DESCEND,
    SAMPLE_E4_GROUND,
    SAMPLE_START,
    SAMPLE_COUNT
};

static char *sSampleFileNames[SAMPLE_COUNT] =
{
    "data/expl0.ogg",
    "data/expl1.ogg",
    "data/expl2.ogg",
    "data/expl3.ogg",
    "data/expl4.ogg",
    "data/expl5.ogg",
    "data/shoot0.ogg",
    "data/shoot1.ogg",
    "data/shoot2.ogg",
    "data/shoot3.ogg",
    "data/hit.ogg",
    "data/end.ogg",
    "data/e2descend.ogg",
    "data/e4ground.ogg",
    "data/powerup.ogg"
};

#define AUDIO_CHANNELS 32

static FMUSIC_MODULE *sMusic;
static FSOUND_SAMPLE *sSamples[SAMPLE_COUNT];


MainState::MainState()
{
}


MainState::~MainState()
{
}


static bool sStringStartsWith(const c8 *str, const c8 *beginning)
{
    irr::core::stringc s(str), b(beginning);
    return s.subString(0, b.size()) == b;
}


bool MainState::init()
{
    Graphics *g = App::getInstance()->getGraphics();

    mTitleUIActive = true;
    mInGame = false;

    if (!FSOUND_Init(44100, AUDIO_CHANNELS, 0))
    {
        gErrorMessage("Audio error", "Unable to initialize sound! (%d)",
                      FSOUND_GetError());
    }
    for (u32 a = 0; a < SAMPLE_COUNT; ++a)
    {
        sSamples[a] = FSOUND_Sample_Load(FSOUND_FREE, sSampleFileNames[a], FSOUND_LOOP_OFF, 0, 0);
        if (!sSamples[a])
            gErrorMessage("Audio error", "Can't load %s", sSampleFileNames[a]);
    }
    sMusic = FMUSIC_LoadSong("data/ld11.it");
    if (!sMusic)
        gErrorMessage("Audio error", "Can't load music");
    FMUSIC_PlaySong(sMusic);

    mFont = new VerticalStripBitmapFontIrr(static_cast<GraphicsIrr *>(g), "data/font.png");
    if (!mFont->init())
        return false;

    // Make sure these are in synch with IMAGE_ID
    mImages.push_back(g->loadImage("data/button.png"));
    mImages.push_back(g->loadImage("data/button-hot.png"));
    mImages.push_back(g->loadImage("data/button-active.png"));
    mImages.push_back(g->loadImage("data/button-hot-active.png"));

    mButtonStyle = new IMGUIImageButtonStyle(mFont, mImages[IMG_BUTTON], mImages[IMG_BUTTON_HOT], mImages[IMG_BUTTON_ACTIVE], mImages[IMG_BUTTON_HOT_ACTIVE]);

    mLabelText = "";

    mDevice = App::getInstance()->getIrrDevice();
    mVideo = App::getInstance()->getIrrVideoDriver();
    mSmgr = App::getInstance()->getIrrSceneManager();


    //mCamera = mSmgr->addCameraSceneNodeMaya(0, -50, 10, 30);
    mCamera = mSmgr->addCameraSceneNode();
    mCamera->setNearValue(0.1f);
    mCamera->setFarValue(10000);
    mCamera->setPosition(core::vector3df(0, 5, -15));
    mCamera->setTarget(core::vector3df(0, 0, 0));
    mCamera->setUpVector(core::vector3df(0, 1, 0));
    mCamera->setFOV((f32)(PI / 2));

    video::ITexture *skyDomeTex = mVideo->getTexture("data/sky.png");
    scene::ISceneNode *skyDomeNode = mSmgr->addSkyDomeSceneNode(skyDomeTex, 20, 8, 1, 2);
    //skyDomeNode->getMaterial(0).setFlag(video::EMF_BILINEAR_FILTER, false);

    SMaterial planeMaterial;
    planeMaterial.Shininess = 0;
    planeMaterial.SpecularColor.set(0);
    planeMaterial.DiffuseColor.set(0);
    planeMaterial.AmbientColor.set(0xffffffff);
    IMesh *planeMesh = mSmgr->addHillPlaneMesh("adf", dimension2df(100, 100), dimension2d<u32>(50, 50), &planeMaterial)->getMesh(0);
    ISceneNode *planeSN = mSmgr->addMeshSceneNode(planeMesh);
    planeSN->setAutomaticCulling(EAC_OFF);

    mSmgr->loadScene("data/scene.irr");

    array<ISceneNode *> nodes;
    u32 a;

    mSmgr->getSceneNodesFromType(ESNT_LIGHT, nodes);
    for (a = 0; a < nodes.size(); ++a)
    {
        ILightSceneNode *lsn = (ILightSceneNode *)nodes[a];
        SLight &ld = lsn->getLightData();
        ld.Attenuation.set(1, 0.035f, 0);
    }
    ISceneNode *lb = mSmgr->getSceneNodeFromName("light_below");
    if (lb)
    {
        SColorf org = ((ILightSceneNode *)lb)->getLightData().DiffuseColor;
        ((ILightSceneNode *)lb)->getLightData().DiffuseColor.set(-org.r, -org.g, -org.b);
    }

    // populate list of route points to correct length first
    mRoutePoints.clear();
    s32 routeLength = mSmgr->getSceneNodeFromName("route_length")->getID();
    for (a = 0; (s32)a < routeLength; ++a)
        mRoutePoints.push_back(RoutePoint());

    // create route from the CP/CT cube nodes
    nodes.clear();
    mSmgr->getSceneNodesFromType(ESNT_CUBE, nodes);
    for (a = 0; a < nodes.size(); ++a)
    {
        ISceneNode *sn = nodes[a];
        if (sStringStartsWith(sn->getName(), "CP"))
        {
            RoutePoint rp;
            c8 name[100];

            s32 parentId = sn->getParent()->getID();
            if (parentId < 0)
                parentId = 0;
            s32 id = sn->getID();

            rp.cameraPos = sn->getAbsolutePosition();

            snprintf(name, 100, "CT%02d", id);
            ISceneNode *target = mSmgr->getSceneNodeFromName(name, sn->getParent());
            if (!target)
            {
                snprintf(name, 100, "CP%02d", id + 1);
                target = mSmgr->getSceneNodeFromName(name, sn->getParent());
            }
            assert(target);
            if (target)
                rp.objPos = target->getAbsolutePosition();

            s32 index = parentId + id;
            mRoutePoints[index] = rp;
        }
    }

    // then remove the CP/CT nodes from the scene
    for (a = 0; a < nodes.size(); ++a)
    {
        ISceneNode *sn = nodes[a];
        const c8 *name = sn->getName();
        if (sStringStartsWith(name, "CP") || sStringStartsWith(name, "CT"))
            sn->remove();
    }

    // read all enemy spawn nodes
    nodes.clear();
    mSmgr->getSceneNodesFromType(ESNT_SPHERE, nodes);
    for (a = 0; a < nodes.size(); ++a)
    {
        ISceneNode *sn = nodes[a];
        const c8 *name = sn->getName();
        if (!sStringStartsWith(name, "Enemy"))
            continue;
        if (!sn->getParent()->isVisible())
        {
            // just remove nodes under hidden parent
            sn->remove();
            continue;
        }

        s32 typeNum = name[5] - '0';
        assert(typeNum >= 0 && typeNum < ET_COUNT);
        if (typeNum < 0 || typeNum >= ET_COUNT)
            continue; // invalid - ignore
        ENEMY_TYPE type = (ENEMY_TYPE)typeNum;
        s32 time = sn->getID() + sn->getParent()->getID();
        assert(time >= -1);

        f32 angle = (sn->getParent()->getRotation().Y + sn->getRotation().Y) * (f32)PI / 180;
        EnemyScriptUnit esu((u32)(time / TIME_SCALE), sn->getAbsolutePosition(), type, -angle);
        mEnemyScript.push_back(esu);

        sn->remove();
    }

    mPlayerParentSceneNode = mSmgr->addEmptySceneNode();

    IAnimatedMesh *playerMesh = mSmgr->getMesh("data/player.obj");
    mPlayerSceneNode = mSmgr->addAnimatedMeshSceneNode(playerMesh, mPlayerParentSceneNode);
    mPlayerSceneNode->getMaterial(0).setFlag(EMF_GOURAUD_SHADING, false);
    mPlayerSceneNode->getMaterial(0).Shininess = 50;
    mPlayerSceneNode->addShadowVolumeSceneNode(0, -1, false);

    IMesh *arrowMesh = mSmgr->getMesh("data/arrow.obj")->getMesh(0);
    mSmgr->getMeshManipulator()->setVertexColors(arrowMesh, SColor(64, 255, 192, 192));
    mArrowSceneNode = mSmgr->addMeshSceneNode(arrowMesh, mPlayerParentSceneNode);
    mArrowSceneNode->getMaterial(0).setFlag(EMF_GOURAUD_SHADING, false);
    mArrowSceneNode->getMaterial(0).setFlag(EMF_LIGHTING, false);
    mArrowSceneNode->getMaterial(0).MaterialType = EMT_ONETEXTURE_BLEND;
    mArrowSceneNode->getMaterial(0).MaterialTypeParam = video::pack_texureBlendFunc(video::EBF_SRC_ALPHA, video::EBF_ONE_MINUS_SRC_ALPHA, video::EMFN_MODULATE_1X);
    mArrowSceneNode->getMaterial(0).DiffuseColor.set(0xff800000);
    mArrowSceneNode->setPosition(vector3df(0, -0.1f, 0));

    mEnemyMeshes[ET_TETRAHEDRON] = mSmgr->getMesh("data/tetrahedron.obj");
    mEnemyMeshes[ET_CUBE] = mSmgr->getMesh("data/cube.obj");
    mEnemyMeshes[ET_DODECAHEDRON] = mSmgr->getMesh("data/dodecahedron.obj");
    mEnemyMeshes[ET_OCTAHEDRON] = mSmgr->getMesh("data/octahedron.obj");
    mEnemyMeshes[ET_ICOSAHEDRON] = mSmgr->getMesh("data/icosahedron.obj");
    mEnemyMeshes[ET_OCTOTOAD] = mSmgr->getMesh("data/octotoad.obj");

    mEnemyShadowMeshes[ET_TETRAHEDRON] = mSmgr->getMesh("data/tetrahedron_shadow.obj");
    mEnemyShadowMeshes[ET_CUBE] = mSmgr->getMesh("data/cube_shadow.obj");
    mEnemyShadowMeshes[ET_DODECAHEDRON] = mSmgr->getMesh("data/dodecahedron_shadow.obj");
    mEnemyShadowMeshes[ET_OCTAHEDRON] = mSmgr->getMesh("data/octahedron_shadow.obj");
    mEnemyShadowMeshes[ET_ICOSAHEDRON] = mSmgr->getMesh("data/icosahedron_shadow.obj");
    mEnemyShadowMeshes[ET_OCTOTOAD] = mSmgr->getMesh("data/octotoad_shadow.obj");

    mBulletMesh = mSmgr->getMesh("data/bullet.obj")->getMesh(0);
    mSmgr->getMeshManipulator()->setVertexColors(mBulletMesh, SColor(32, 16, 0, 0));
    mBulletMaterial.setFlag(EMF_GOURAUD_SHADING, false);
    mBulletMaterial.setFlag(EMF_LIGHTING, false);
    mBulletMaterial.MaterialType = EMT_TRANSPARENT_VERTEX_ALPHA;
    mBulletMaterial.DiffuseColor.set(0xffffffff);

    IMesh *cylinderPipeShadowMesh = mSmgr->getMesh("data/cylinder_pipe_shadow.obj");

    nodes.clear();
    mSmgr->getSceneNodesFromType(ESNT_ANIMATED_MESH, nodes);
    for (a = 0; a < nodes.size(); ++a)
    {
        ISceneNode *node = nodes[a];
        assert(node->getType() == ESNT_ANIMATED_MESH);
        IAnimatedMeshSceneNode *anode = (IAnimatedMeshSceneNode *)node;
        if (sStringStartsWith(node->getName(), "cube"))
            anode->addShadowVolumeSceneNode(mEnemyShadowMeshes[ET_CUBE], -1, false);
        else if (sStringStartsWith(node->getName(), "tetrahedron"))
            anode->addShadowVolumeSceneNode(mEnemyShadowMeshes[ET_TETRAHEDRON], -1, false);
        else if (sStringStartsWith(node->getName(), "icosahedron"))
            anode->addShadowVolumeSceneNode(mEnemyShadowMeshes[ET_ICOSAHEDRON], -1, false);
        else if (sStringStartsWith(node->getName(), "octahedron"))
            anode->addShadowVolumeSceneNode(mEnemyShadowMeshes[ET_OCTAHEDRON], -1, false);
        else if (sStringStartsWith(node->getName(), "octotoad"))
            anode->addShadowVolumeSceneNode(mEnemyShadowMeshes[ET_OCTOTOAD], -1, false);
        else if (sStringStartsWith(node->getName(), "dodecahedron"))
            anode->addShadowVolumeSceneNode(mEnemyShadowMeshes[ET_DODECAHEDRON], -1, false);
        else if (sStringStartsWith(node->getName(), "cylinder_pipe"))
            anode->addShadowVolumeSceneNode(cylinderPipeShadowMesh, -1, false);
        else if (sStringStartsWith(node->getName(), "bcube"))
            anode->addShadowVolumeSceneNode(0, -1, false);

        anode->getMaterial(0).FogEnable = true;
        anode->setAutomaticCulling(EAC_OFF);
    }

    mSmgr->setShadowColor(SColor(0x14001040));

    mVideo->setFog(SColor(0, 198, 201, 208), true, 15, 50, 0.1f);

    mAimAngle = 0;
    mLastMousePos.set(WIDTH / 2, HEIGHT / 2);

    /*
    mEnemyScript.push_back(EnemyScriptUnit(3000, vector3df(-3.121577f, 1.666666f, -19.066460f), ET_CUBE));
    mEnemyScript.push_back(EnemyScriptUnit(2000, vector3df(1.692073f, 1.666666f, -24.203796f), ET_TETRAHEDRON));
    */
    mCurrentEnemyScriptIndex = 0;
    mScore = 0;

    std::sort(mEnemyScript.begin(), mEnemyScript.end());

    // some weird pos for title screen
    mPlayerParentSceneNode->setPosition(vector3df(-100, 1, -100));

    return mFont != NULL && mButtonStyle != NULL;
}


void MainState::deinit()
{
    UI32 a;
    delete mFont;
    for (a = 0; a < mImages.size(); ++a)
        delete mImages[a];
    mImages.clear();

    FMUSIC_StopAllSongs();
    FMUSIC_FreeSong(sMusic);
    for (a = 0; a < AUDIO_CHANNELS; ++a)
        FSOUND_StopSound(a);
    for (a = 0; a < SAMPLE_COUNT; ++a)
        FSOUND_Sample_Free(sSamples[a]);
    FSOUND_Close();
}


void MainState::onActivate(UI32 time)
{
    AppState::onActivate(time);
    mStartTime = 0; // first call to update will do actual initialization
    mPaused = false;
}


bool MainState::onKeyUp(const irr::SEvent::SKeyInput & /*event*/)
{
    return false;
}

bool MainState::onKeyDown(const irr::SEvent::SKeyInput &event)
{
    if (event.Key == KEY_ESCAPE)
        App::getInstance()->setAlive(false);

    return false;
}

bool MainState::onMouseDown(const irr::SEvent::SMouseInput & /*event*/)
{
    if (mTitleUIActive || mEndingUIActive)
        mIMGUI.onMouseDown();

    if (mInGame)
    {
        Bullet b;
        b.pos = mPlayerParentSceneNode->getPosition();
        b.framesLeft = 180;
        ISceneNode *sn = mSmgr->addMeshSceneNode(mBulletMesh);
        sn->getMaterial(0) = mBulletMaterial;
        sn->setPosition(b.pos);
        sn->setRotation(vector3df(0, (f32)(mAimAngle * 180 / PI + 180), 0));
        b.speed.set(sin(mAimAngle), 0, cos(mAimAngle));
        b.speed /= 3.f;
        b.sceneNode = sn;
        mBullets.push_back(b);
        FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_SHOOT_FIRST + rand() % SAMPLE_SHOOT_COUNT]);
    }

    return false;
}

bool MainState::onMouseUp(const irr::SEvent::SMouseInput & /*event*/)
{
    mIMGUI.onMouseUp();
    return false;
}

bool MainState::onMouseMotion(const irr::SEvent::SMouseInput &event)
{
    mLastMousePos.set(event.X, event.Y);
    mIMGUI.onMouseMotion(event.X, event.Y);
    return false;
}

bool MainState::onMouseWheel(const irr::SEvent::SMouseInput & /*event*/)
{
    return false;
}


static s32 sClampSplineIndex(s32 idx, s32 size)
{
	//return ( idx<0 ? size+idx : ( idx>=size ? idx-size : idx ) );  // loop
    if (idx < 0)
        idx = 0;
    //if (idx >= size)
    //    idx = size - 1;
    if (idx >= size)
        idx = size - 3 + idx % 3; // loop last 3 control points
    return idx;
}

// Spline interpolation code is stolen from Irr's CSceneNodeAnimatorFollowSpline
vector3df MainState::getSplinePosition(u32 time, f32 speed, f32 tightness, bool target)
{
	const u32 pSize = (u32)mRoutePoints.size();
	if (pSize == 0)
		return vector3df(0, 0, 0);

	const f32 dt = (time / 1000.f) * speed;
	const f32 u = core::fract(dt);
	const s32 idx = core::floor32(dt); // % pSize; - disabled wrap-around...
	
    s32 idx0 = sClampSplineIndex(idx - 1, pSize);
    s32 idx1 = sClampSplineIndex(idx, pSize);
    s32 idx2 = sClampSplineIndex(idx + 1, pSize);
    s32 idx3 = sClampSplineIndex(idx + 2, pSize);
    const core::vector3df &p0 = target ? mRoutePoints[idx0].objPos : mRoutePoints[idx0].cameraPos;
	const core::vector3df &p1 = target ? mRoutePoints[idx1].objPos : mRoutePoints[idx1].cameraPos;
	const core::vector3df &p2 = target ? mRoutePoints[idx2].objPos : mRoutePoints[idx2].cameraPos;
	const core::vector3df &p3 = target ? mRoutePoints[idx3].objPos : mRoutePoints[idx3].cameraPos;

	// hermite polynomials
	const f32 h1 = 2.0f * u * u * u - 3.0f * u * u + 1.0f;
	const f32 h2 = -2.0f * u * u * u + 3.0f * u * u;
	const f32 h3 = u * u * u - 2.0f * u * u + u;
	const f32 h4 = u * u * u - u * u;

	// tangents
	const core::vector3df t1 = ( p2 - p0 ) * tightness;
	const core::vector3df t2 = ( p3 - p1 ) * tightness;

	// interpolated point
    return p1 * h1 + p2 * h2 + t1 * h3 + t2 * h4;
}


void MainState::runEnemyScriptSpawns()
{
    if (mCurrentEnemyScriptIndex >= mEnemyScript.size())
        return; // the level should be finished by now...
    EnemyScriptUnit &esu = mEnemyScript[mCurrentEnemyScriptIndex];
    if (mTime >= esu.time)
    {
        IAnimatedMeshSceneNode *sn = mSmgr->addAnimatedMeshSceneNode(mEnemyMeshes[esu.type]);
        if (mEnemyShadowMeshes[esu.type])
            sn->addShadowVolumeSceneNode(mEnemyShadowMeshes[esu.type], -1, false);
        //esu.pos.Y = -2.f;
        sn->setPosition(esu.pos);

        Enemy e(esu.type, esu.pos, esu.angle, sn, sEnemyRadii[esu.type]);
        mEnemies.push_back(e);

        ++mCurrentEnemyScriptIndex;
    }
}


void MainState::killEnemy(s32 index, bool getScore, bool oldAge)
{
    Enemy &e = mEnemies[index];
    if (e.deathCount >= 0)
        return; // already dying
    e.deathCount = ENEMY_DEATH_DURATION;
    e.sceneNode->setMesh(mEnemyShadowMeshes[e.type]);
    e.sceneNode->getMaterial(0).setFlag(EMF_GOURAUD_SHADING, false);

    if (!oldAge)
    {
        if (getScore)
        {
            FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_EXPLOSION_FIRST + e.type]);
            mScore += e.type + 1;
        }
        else
        {
            FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_HIT]);
            mScore -= e.type / 2 + 1;
        }
    }
}


void MainState::moveTowardsPlayer(vector3df &ioPos, f32 speed)
{
    vector3df dir = mPlayerParentSceneNode->getPosition() - ioPos;
    dir.normalize();
    dir.Y = 0;
    ioPos += dir * speed;
}


void MainState::updateEnemies()
{
    u32 a;
    for (a = 0; a < mEnemies.size(); )
    {
        bool removed = false;

        Enemy &e = mEnemies[a];

        switch (e.type)
        {
        case ET_TETRAHEDRON:
            e.pos.X += cos(e.angle) * 0.1f;
            e.pos.Y = e.r * 1.2f;
            e.pos.Z += sin(e.angle) * 0.1f;
            if (e.age < 30)
            {
                f32 scale = e.age / 30.f;
                e.sceneNode->setScale(vector3df(scale, scale, scale));
            }
            break;
        case ET_CUBE:
            {
                moveTowardsPlayer(e.pos, 0.04f);
                e.pos.Y = sinf(fmodf(mTime * 0.003f, (f32)PI)) + e.r - 0.45f;
                if (e.age < 50)
                    e.pos.Y += 2 * (-e.r + e.r * e.age / 50);
            }
            break;
        case ET_DODECAHEDRON:
            if (e.age < 200)
            {
                if (e.age == 1)
                    FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_E2_DESCEND]);
                float k = 1 - sinf(e.age * (f32)PI / (2 * 200));
                e.pos.Y = 50 * (k * k) + e.r;
            }
            e.angle += 0.02f;
            moveTowardsPlayer(e.pos, 0.06f);
            break;
        case ET_OCTAHEDRON:
            if (e.age < 100)
                e.pos.Y = 3 * (-e.r + e.r * e.age / 100) + e.r * 1.1f;
            if ((mTime / 500) % 3 == 0)
                moveTowardsPlayer(e.pos, 0.2f);
            break;
        case ET_ICOSAHEDRON:
            if (e.age < 100)
            {
                float k = cosf(e.age * (f32)PI / (2 * 100));
                e.pos.Y = 50 * (k) + e.r;
            }
            if (e.age >= 100 && e.age <= 130)
            {
                if (e.age == 100)
                    FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_E4_GROUND]);
                float m = (130 - e.age) / 30.f;
                e.pos.Y = e.r + sinf(e.age * 2.f) * m * 0.1f;
            }
            if ((mTime / 500) % 2 == 0)
                moveTowardsPlayer(e.pos, 0.15f);
            break;
        case ET_OCTOTOAD:
            {
                if (e.age < 100)
                    e.pos.Y = 5 * (-e.r + e.r * e.age / 100) + e.r * 1.3f;
                f32 scale = 1.5f - sinf(fmodf(mTime * 0.003f, (f32)PI)) * 0.5f;
                e.sceneNode->setScale(vector3df(scale, scale, scale));
                moveTowardsPlayer(e.pos, 0.1f);
            }
            break;
        }

        e.sceneNode->setPosition(e.pos);
        e.sceneNode->setRotation(vector3df(0, e.angle * 180 / (f32)PI, 0));

        f32 minDist = PLAYER_RADIUS + e.r;
        if (e.pos.getDistanceFromSQ(mPlayerParentSceneNode->getPosition()) < minDist * minDist)
        {
            // enemy collides with the player
            //e.sceneNode->remove();
            killEnemy(a, false, false);
        }

        if (++e.age >= sEnemyAges[e.type])
            killEnemy(a, false, true);

        if (e.deathCount >= 0)
        {
            f32 scale = (f32)e.deathCount / ENEMY_DEATH_DURATION;
            e.sceneNode->setScale(vector3df(scale, scale, scale));
            if (e.deathCount-- == 0)
            {
                e.sceneNode->remove();
                mEnemies.erase(mEnemies.begin() + a);
                removed = true;
            }
        }

        if (!removed)
            ++a;
    }
}


void MainState::updateBullets()
{
    u32 a;
    for (a = 0; a < mBullets.size(); )
    {
        bool removeBullet = false;
        Bullet &b = mBullets[a];

        b.pos += b.speed;
        b.sceneNode->setPosition(b.pos);

        for (u32 ei = 0; ei < mEnemies.size(); ++ei)
        {
            Enemy &e = mEnemies[ei];
            if (e.deathCount >= 0)
                continue; // enemy is already dying
            f32 minDist = BULLET_RADIUS + e.r;
            if (e.pos.getDistanceFromSQ(b.pos) < minDist * minDist)
            {
                killEnemy(ei, true, false);
                removeBullet = true;
                break;
            }
        }

        if (removeBullet || --b.framesLeft <= 0)
        {
            b.sceneNode->remove();
            mBullets.erase(mBullets.begin() + a);
        }
        else
            ++a;
    }
}


bool MainState::update(UI32 time)
{
    /*
    // for testing - skip time
    if (mStartTime == 0 && mInGame)
    {
        int skip = (int)(55000 / TIME_SCALE);
        mStartTime = time - skip;
        while (mCurrentEnemyScriptIndex < mEnemyScript.size() &&
               (s32)mEnemyScript[mCurrentEnemyScriptIndex].time < skip)
            ++mCurrentEnemyScriptIndex;
    }
    */

    bool result = AppState::update(time);
    if (!result)
        return false; // game is paused...

    if (mTitleUIActive)
    {
        vector3df pos = mPlayerParentSceneNode->getPosition();
        pos.Y -= 1.f;
        mCamera->setTarget(pos);
        pos.X += sinf(time * 0.0005f) * 3;
        pos.Z += cosf(time * 0.0005f) * 2;
        pos.Y += 2.f;
        mCamera->setPosition(pos);

        return false;
    }

    if (mInGame && mTime >= mRoutePoints.size() * 1000 / TIME_SCALE)
    {
        mInGame = false;
        mEndingUIActive = true;
        FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_END]);
    }

    // Following is executed only when game is not paused & game is running.

    runEnemyScriptSpawns();
    updateEnemies();
    updateBullets();


    const f32 speed = TIME_SCALE, tightness = 0.5f;
    vector3df camPos = getSplinePosition(mTime, speed, tightness);
    camPos.Y += 1.5f;
    mCamera->setPosition(camPos);
    vector3df pos = getSplinePosition(mTime, speed, tightness, true);
    mCamera->setTarget(pos);

    vector3df prevPos = mPlayerParentSceneNode->getPosition();
    pos.Y = 0.8f;
    mPlayerParentSceneNode->setPosition(pos);
    f32 angle = atan2(pos.X - prevPos.X, pos.Z - prevPos.Z);
    mPlayerSceneNode->setRotation(vector3df(mTime * 0.2f, (f32)(angle * 180 / PI), 0));

    mAimAngle = atan2((f32)(mLastMousePos.Y - HEIGHT * 18 / 30), (f32)(mLastMousePos.X - WIDTH / 2));
    vector3df camVec = pos - camPos;
    mAimAngle += (f32)(atan2(camVec.X, camVec.Z) + PI / 2);
    mArrowSceneNode->setRotation(vector3df(0, (f32)(mAimAngle * 180 / PI), 0));

    return true;
}


void MainState::render(irr::IrrlichtDevice *device)
{
    assert(device);
    irr::video::IVideoDriver *driver = device->getVideoDriver();
    assert(driver);
    irr::scene::ISceneManager *smgr = device->getSceneManager();

    driver->beginScene(true, true, video::SColor(0x00445566));

    smgr->drawAll();

    Graphics *g = App::getInstance()->getGraphics();
    g->beginRender();
    mIMGUI.begin();

    if (mTitleUIActive)
    {
        mFont->drawString(WIDTH / 2, 150, "warrior", FA_HORIZ_CENTER, Color(0xff000000));
        mFont->drawString(WIDTH / 2, 180, "a casual railshooter", FA_HORIZ_CENTER, 0.7f, Color(0x80000000));

        mFont->drawString(WIDTH / 2, 250, "move mouse to aim", FA_HORIZ_CENTER, 0.7f, Color(0xaa000000));
        mFont->drawString(WIDTH / 2, 280, "click button to shoot", FA_HORIZ_CENTER, 0.7f, Color(0xaa000000));

        if (mIMGUI.button(WIDTH / 2 - mButtonStyle->getWidth() / 2, 350, IMGUI_AI_COUNTER, mButtonStyle, 0, 0, "Play"))
        {
            mTitleUIActive = false;
            mInGame = true;
            mStartTime = 0;
            FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_START]);
        }
        if (mIMGUI.button(WIDTH / 2 - mButtonStyle->getWidth() / 2, 450, IMGUI_AI_COUNTER, mButtonStyle, 0, 0, "Quit"))
            App::getInstance()->setAlive(false);

        mFont->drawString(WIDTH / 2, HEIGHT - 5, "Copyright 2008 Jetro Lauha - www.jet.ro", FA_HORIZ_CENTER | FA_BOTTOM, 0.5f, Color(0x80000000));
    }
    else if (mInGame)
    {
        char buf[100];
        sprintf(buf, "%d", mScore);
        mFont->drawString(40, 40, "score", FA_DEFAULT, 0.7f, Color(0x88000000));
        mFont->drawString(40, 60, buf, FA_DEFAULT, Color(0xff000000));
    }
    else if (mEndingUIActive)
    {
        char *title = "";
        char buf[100];
        sprintf(buf, "%d", mScore);
        if (mScore < 0)
            title = "Failboat!";
        else if (mScore == 0)
            title = "No pain, no gain. Game over!";
        else if (mScore > 0 && mScore < 40)
            title = "Game over.";
        else if (mScore >= 40 && mScore < 100)
            title = "Good job!";
        else if (mScore >= 100 && mScore < 300)
            title = "Success!";
        else if (mScore >= 300)
            title = "Awesome!";
        mFont->drawString(WIDTH / 2, 200, title, FA_HORIZ_CENTER, Color(0xff000000));
        mFont->drawString(WIDTH / 2, 300, "Final score:", FA_HORIZ_CENTER, 0.6f, Color(0x80000000));
        mFont->drawString(WIDTH / 2, 350, buf, FA_HORIZ_CENTER, 0.8f, Color(0xcc000000));
        if (mIMGUI.button(WIDTH / 2 - mButtonStyle->getWidth() / 2, 450, IMGUI_AI_COUNTER, mButtonStyle, 0, 0, "Quit"))
            App::getInstance()->setAlive(false);
    }

    mIMGUI.end();
    g->endRender();

    driver->endScene();
}
