/* Turska Framework/UI library (Turs2)
 * Copyright 2003-2008 Jetro Lauha
 * All rights reserved.
 * Web: http://iki.fi/jetro/turska/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of EITHER:
 *   (1) The GNU Lesser General Public License as published by the Free
 *       Software Foundation; either version 2.1 of the License, or (at
 *       your option) any later version. The text of the GNU Lesser
 *       General Public License is included with this library in the
 *       file LICENSE-LGPL.txt.
 *   (2) The BSD-style license that is included with this library in
 *       the file LICENSE-BSD.txt.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the files
 * LICENSE-LGPL.txt and LICENSE-BSD.txt for more details.
 *
 * $Id: T2Font.h,v 1.1 2008-04-19 11:46:25 tonic Exp $
 * $Revision: 1.1 $
 */

#ifndef T2FONT_H_INCLUDED
#define T2FONT_H_INCLUDED


namespace turska
{

    //! Font alignment flags.
    /*!
     * Default alignment value of 0 implies top vertical alignment and
     * left horizontal alignment. Horizontal alignment flags are
     * exclusive with each other. Similarly also vertical alignment flags
     * are exclusive with each other.
     */
    enum FONT_ALIGN_FLAG
    {
        FA_LEFT         = 0,
        FA_HORIZ_CENTER = (1 << 0),
        FA_RIGHT        = (1 << 1),
        FA_TOP          = 0,
        FA_VERT_CENTER  = (1 << 2),
        FA_BOTTOM       = (1 << 3),
        FA_DEFAULT      = FA_LEFT | FA_TOP
    };

    //! Interface for font class implementations.
    /*! \note Current API and implementation details are subject to changes
     *        which may break backwards compatibility.
     * \todo Support for proper font metrics (baseline, ascent, descent, ...).
     * \todo Support for multi-line texts, measurement for amount of lines, ...
     */
    class Font
    {
    public:

        Font() : mLoaded(false), mDrawMode(IDM_NORMAL) {}
        virtual ~Font() {}

        //! Does any initialization the object needs before it is ready for use.
        /*!
         * \return TRUE if successful or FALSE otherwise.
         */
        virtual bool init() = 0;

        //! Returns TRUE if the font is loaded or FALSE otherwise.
        bool isLoaded() { return mLoaded; }

        //! Sets horizontal character spacing.
        virtual void setCharacterSpacing(I32 spacing) = 0;
        //! Returns horizontal character spacing.
        virtual I32 getCharacterSpacing() = 0;
        //! Returns height of the font.
        virtual UI32 getFontHeight() = 0;

        //! Reloads the font.
        /*!
         * Sets the internal loaded state to true if loading was successful
         * and FALSE otherwise.
         * \return true if successful or false otherwise.
         */
        virtual bool reload() = 0;

        //! Returns width of given string using current font.
        virtual I32 getStringWidth(const C8 *string) = 0;

        //! Returns width of given string in given scale using current font.
        virtual REAL getStringWidth(const C8 *string, REAL scale) = 0;

        //! Sets draw mode for glyphs. \sa IMAGE_DRAW_MODE
        void setDrawMode(IMAGE_DRAW_MODE drawMode) { mDrawMode = drawMode; }

        //! Draws a string.
        /*!
         * \param x X-coordinate of the origin for the string.
         * \param y Y-coordinate of the origin for the string.
         * \param string string to draw.
         * \param flags flags to control how the string is positioned and
         *              drawn. \sa FONT_ALIGN
         * \param color color and alpha to use.
         */
        virtual void drawString(I32 x, I32 y, const C8 *string,
                                UI32 flags = FA_DEFAULT,
                                const Color &color = Color(0xffffffff)) = 0;

        //! Draws a string.
        /*!
         * \param x X-coordinate of the origin for the string.
         * \param y Y-coordinate of the origin for the string.
         * \param string string to draw.
         * \param flags flags to control how the string is positioned and
         *              drawn. \sa FONT_ALIGN
         * \param scale relative scale to draw with.
         * \param color color and alpha to use.
         */
        virtual void drawString(REAL x, REAL y, const C8 *string,
                                UI32 flags = FA_DEFAULT, REAL scale = 1,
                                const Color &color = Color(0xffffffff)) = 0;

    protected:

        bool mLoaded;
        IMAGE_DRAW_MODE mDrawMode;

    }; // class Font

} // namespace turska


#endif // !T2FONT_H_INCLUDED
