/* Turska Framework/UI library (Turs2)
 * Copyright 2003-2008 Jetro Lauha
 * All rights reserved.
 * Web: http://iki.fi/jetro/turska/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of EITHER:
 *   (1) The GNU Lesser General Public License as published by the Free
 *       Software Foundation; either version 2.1 of the License, or (at
 *       your option) any later version. The text of the GNU Lesser
 *       General Public License is included with this library in the
 *       file LICENSE-LGPL.txt.
 *   (2) The BSD-style license that is included with this library in
 *       the file LICENSE-BSD.txt.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the files
 * LICENSE-LGPL.txt and LICENSE-BSD.txt for more details.
 *
 * $Id: T2Image.h,v 1.1 2008-04-19 11:46:25 tonic Exp $
 * $Revision: 1.1 $
 */

#ifndef T2IMAGE_H_INCLUDED
#define T2IMAGE_H_INCLUDED


namespace turska
{

    //! Image drawing modes.
    enum IMAGE_DRAW_MODE
    {
        IDM_NORMAL = 0, //!< Normal image drawing, with alpha if supported.
        IDM_ADD,        //!< Additive image drawing.
        IDM_MULTIPLY,   //!< Multiplicative image drawing.
        //! Count of image drawing modes.
        IDM_DRAW_MODE_COUNT // Do not add new draw modes after this one.
    };

    //! Drawing flags for image features to use when drawing.
    /*! Note that implementations are free to ignore selected flags.
     * For example, a software rendered implementation might ignore
     * rotation. Or a simple hardware rendered implementation might
     * ignore color and alpha flags and assume those are always used.
     */
    enum IMAGE_FLAG
    {
        //! Multiplies source image with color when drawing.
        IF_COLOR         = (1 << 0),
        //! Multiplies source image with alpha when drawing.
        IF_ALPHA         = (1 << 1),
        
        /*! Enables transformation. Drawing uses just target width & height
         *  when disabled and ROTATE, CENTER and SCALE flags when enabled.
         */
        IF_TRANSFORM     = (1 << 8),
        //! Rotates image when drawing.
        IF_ROTATE        = (1 << 9),
        //! Uses center or given pivot of the image as origin instead of top-left corner.
        IF_CENTER        = (1 << 10),
        //! Applies scaling to image. See setScale() methods.
        IF_SCALE         = (1 << 11),
        
        //! Uses bilinear filtering for image sampling if available.
        /*! Recommended for use when images will be scaled.
         * Default is to use nearest neighbor sampling.
         * Currently this flag can be set only when loading images.
         */
        IF_FILTER_LINEAR = (1 << 16),
        
        //! Configures image wrap mode to repeat. Default is to clamp.
        /*! Currently this flag can be set only when loading images.
         */
        IF_WRAP_REPEAT   = (1 << 17),
        
        /*! Default set of flags for image.
         * \note The IF_CENTER flag alone doesn't mean all images will be
         *       centered. The IF_CENTER flag affects only when IF_TRANSFORM
         *       flag is enabled in addition.
         */
        IF_DEFAULT_FLAGS = IF_COLOR | IF_CENTER
    };
    
    
    //! Abstract interface for a container of a static UI image.
    /*! \note Implementation of this class may not support all
     *        features defined by the interface.
     */
    class Image
    {
    public:
        
        //! Initializes image with given values.
        /*!
         * \param width width of the image.
         * \param height height of the image.
         * \param drawMode initial drawing mode to use, default is IDM_NORMAL.
         * \param flags initial flags to use. \sa IF_DEFAULT_FLAGS
         */
        Image(UI32 width, UI32 height,
              IMAGE_DRAW_MODE drawMode = IDM_NORMAL,
              UI32 flags = IF_DEFAULT_FLAGS) :
            mWidth(width), mHeight(height), mFlags(flags), mDrawMode(drawMode) {}
        
        virtual ~Image() {}

        
        inline UI32 getWidth() { return mWidth; }
        inline UI32 getHeight() { return mHeight; }

        inline void setFlags(UI32 flags) { mFlags = flags; }
        inline UI32 getFlags() const { return mFlags; }
        inline void setFlag(UI32 flag) { mFlags |= flag; }
        inline void clearFlag(UI32 flag) { mFlags &= ~flag; }
        inline void setFlagState(UI32 flag, bool enabled)
            { enabled ? setFlag(flag) : clearFlag(flag); }
        inline bool isFlag(UI32 flag) const { return (mFlags & flag) == flag; }

        //! Sets drawing mode. \sa IMAGE_DRAW_MODE
        inline void setDrawMode(IMAGE_DRAW_MODE drawMode) { mDrawMode = drawMode; }
        
        //! Returns current drawing mode. \sa IMAGE_DRAW_MODE
        inline IMAGE_DRAW_MODE getDrawMode() { return mDrawMode; }

        
        //! Sets the color and alpha to use when IF_COLOR or IF_ALPHA flags are enabled.
        virtual void setColor(const Color &color) = 0;
        
        
        //! Sets target rectangle size to use when TRANSFORM flag is disabled.
        /*! \note Value scale is the display unit size in use, typically pixels.
         */
        virtual void setTargetSize(UI32 width, UI32 height) = 0;
        
        //! Sets target rectangle size to use when IF_TRANSFORM flag is disabled.
        /*! \note Value scale is the display unit size in use, typically pixels.
         */
        virtual void setTargetSize(REAL width, REAL height) = 0;
        

        //! Sets rotation angle to use when IF_ROTATE flag is enabled.
        /*! \param angle rotation angle in radians.
         */
        virtual void setAngle(REAL angle) = 0;
        

        //! Sets the scaling factor to use when SCALE flag is enabled.
        virtual void setScale(REAL scale) = 0;

        //! Sets non-uniform scaling factors to use when SCALE flag is enabled.
        virtual void setScale(REAL scaleX, REAL scaleY) = 0;

        
        //! Sets pivot to use when CENTER flag is enabled. Default is the center of image.
        virtual void setPivot(REAL centerX, REAL centerY) = 0;


        //! Draws image using current flags, with reference point (x,y).
        virtual void draw(I32 x, I32 y) = 0;
        
        //! Draws image using current flags, with reference point (x,y).
        virtual void draw(REAL x, REAL y) = 0;
        

        //! Draws image using given drawing mode and flags.
        virtual void draw(IMAGE_DRAW_MODE drawMode, UI32 flags,
                          I32 x, I32 y) = 0;
        
        //! Draws image using given drawing mode and flags.
        virtual void draw(IMAGE_DRAW_MODE drawMode, UI32 flags,
                          REAL x, REAL y) = 0;


        //! Draws image using given drawing mode, flags, color and alpha.
        virtual void draw(IMAGE_DRAW_MODE drawMode, UI32 flags,
                          REAL x, REAL y, const Color &color) = 0;

    protected:

        UI32 mWidth;      //!< Image width.
        UI32 mHeight;     //!< Image height.
        UI32 mFlags;      //!< Current flag state. \sa FLAGS
        IMAGE_DRAW_MODE mDrawMode; //!< Current drawing mode. \sa DRAW_MODE
        
    }; // class Image

} // namespace turska


#endif // !T2IMAGE_H_INCLUDED
