/* Turska Framework/UI library (Turs2)
 * Copyright 2003-2008 Jetro Lauha
 * All rights reserved.
 * Web: http://iki.fi/jetro/turska/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of EITHER:
 *   (1) The GNU Lesser General Public License as published by the Free
 *       Software Foundation; either version 2.1 of the License, or (at
 *       your option) any later version. The text of the GNU Lesser
 *       General Public License is included with this library in the
 *       file LICENSE-LGPL.txt.
 *   (2) The BSD-style license that is included with this library in
 *       the file LICENSE-BSD.txt.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the files
 * LICENSE-LGPL.txt and LICENSE-BSD.txt for more details.
 *
 * $Id: T2IMGUI.cpp,v 1.1 2008-04-19 11:46:25 tonic Exp $
 * $Revision: 1.1 $
 */

#include "Turs2.h"


namespace turska
{


IMGUI::IMGUI() :
    mMouseX(0), mMouseY(0),
    mMouseDown(false),
    mHotItemId(IMGUI_AWS_NONE),
    mActiveItemId(IMGUI_AWS_NONE),
    mAutoIdCounter(0)
{
}


IMGUI::~IMGUI()
{
}


void IMGUI::begin(I16 groupId)
{
    mHotItemId = IMGUI_AWS_NONE;
    mAutoIdCounter = (groupId & 0x7fff) << 16;
}


void IMGUI::end()
{
    if (!mMouseDown)
        mActiveItemId = IMGUI_AWS_NONE;
    else if (mActiveItemId == IMGUI_AWS_NONE)
        mActiveItemId = IMGUI_AWS_UNAVAILABLE; // this isn't currently used for anything
}


I32 IMGUI::defaultWidgetStateLogic(I32 id, bool hotNow)
{
    if (hotNow)
    {
        mHotItemId = id;
        if (mActiveItemId == IMGUI_AWS_NONE && mMouseDown)
            mActiveItemId = id;
    }

    I32 widgetState = IMGUI_WS_DEFAULT;
    if (mHotItemId == id)
    {
        if (mActiveItemId == id)
            widgetState = IMGUI_WS_HOT_ACTIVE;
        else
            widgetState = IMGUI_WS_HOT;
    }
    else if (mActiveItemId == id)
        widgetState = IMGUI_WS_ACTIVE;

    return widgetState;
}


I32 IMGUI::buttonLogic(I32 id, IMGUIButtonStyle *buttonStyle,
                       REAL x, REAL y, REAL width, REAL height,
                       const C8 *text, Image *image)
{
    bool inside = false;
    if (buttonStyle != NULL)
        inside = buttonStyle->isInside(mMouseX - x, mMouseY - y, width, height, text, image);

    I32 buttonState = defaultWidgetStateLogic(id, inside);

    return buttonState;
}


bool IMGUI::isButtonClicked(I32 id)
{
    bool clicked = false;

    if (!mMouseDown && mHotItemId == id && mActiveItemId == id)
        clicked = true;

    return clicked;
}


bool IMGUI::button(REAL x, REAL y, I32 id,
                   IMGUIButtonStyle *buttonStyle,
                   REAL width, REAL height,
                   const C8 *text, Image *image)
{
    if (id == IMGUI_AI_COUNTER)
        id = ++mAutoIdCounter;
    
    I32 buttonState = buttonLogic(id, buttonStyle, x, y, width, height, text, image);
    
    buttonStyle->render(buttonState, x, y, width, height, text, image);

    return isButtonClicked(id);
}


void IMGUI::slider(REAL x, REAL y, IMGUI_WIDGET_ORIENTATION axis,
                   I32 &value, I32 minValue, I32 maxValue,
                   I32 id, IMGUISliderStyle *sliderStyle,
                   REAL width, REAL height)
{
    if (id == IMGUI_AI_COUNTER)
        id = ++mAutoIdCounter;
    
    bool inside = false;
    if (sliderStyle != NULL)
        inside = sliderStyle->isInside(mMouseX - x, mMouseY - y, width, height);

    I32 sliderState = defaultWidgetStateLogic(id, inside);

    if (sliderStyle != NULL && 
        (sliderState & IMGUI_WS_MASK_ACTIVE) != 0)
    {
        I32 newValue = sliderStyle->getIntValue(axis, minValue, maxValue,
                                                mMouseX - x, mMouseY - y, width, height);
        value = newValue;
    }

    sliderStyle->render(sliderState, axis, value, minValue, maxValue, x, y, width, height);
}


void IMGUI::slider(REAL x, REAL y, IMGUI_WIDGET_ORIENTATION axis,
                   REAL &value, REAL minValue, REAL maxValue,
                   I32 id, IMGUISliderStyle *sliderStyle,
                   REAL width, REAL height)
{
    if (id == IMGUI_AI_COUNTER)
        id = ++mAutoIdCounter;

    bool inside = false;
    if (sliderStyle != NULL)
        inside = sliderStyle->isInside(mMouseX - x, mMouseY - y, width, height);

    I32 sliderState = defaultWidgetStateLogic(id, inside);

    if (sliderStyle != NULL && 
        (sliderState & IMGUI_WS_MASK_ACTIVE) != 0)
    {
        REAL newValue = sliderStyle->getRealValue(axis, minValue, maxValue,
                                                  mMouseX - x, mMouseY - y, width, height);
        value = newValue;
    }

    sliderStyle->render(sliderState, axis, value, minValue, maxValue, x, y, width, height);
}


void IMGUI::onKeyDown(const KeyEvent &key)
{
    // TODO - keyboard usage...
}


void IMGUI::onMouseDown()
{
    mMouseDown = true;
}


void IMGUI::onMouseUp()
{
    mMouseDown = false;
}


void IMGUI::onMouseMotion(I32 x, I32 y)
{
    mMouseX = x;
    mMouseY = y;
}


} // namespace turska
