/* Turska Framework/UI library (Turs2)
 * Copyright 2003-2008 Jetro Lauha
 * All rights reserved.
 * Web: http://iki.fi/jetro/turska/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of EITHER:
 *   (1) The GNU Lesser General Public License as published by the Free
 *       Software Foundation; either version 2.1 of the License, or (at
 *       your option) any later version. The text of the GNU Lesser
 *       General Public License is included with this library in the
 *       file LICENSE-LGPL.txt.
 *   (2) The BSD-style license that is included with this library in
 *       the file LICENSE-BSD.txt.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the files
 * LICENSE-LGPL.txt and LICENSE-BSD.txt for more details.
 *
 * $Id: T2ImageIrr.cpp,v 1.1 2008-04-19 11:46:25 tonic Exp $
 * $Revision: 1.1 $
 */

#ifdef T2_USE_IRR

#include <assert.h>
#include "irrlicht.h"

#include "Turs2.h"

using namespace irr;

namespace turska
{


ImageIrr::ImageIrr(GraphicsIrr *graphics,
                   UI32 width, UI32 height, UI32 flags,
                   irr::video::ITexture *texture, const irr::core::vector2df *uv) :
    Image(width, height, IDM_NORMAL, flags),
    mGraphics(graphics),
    mDriver(0),
    mTexture(texture),
    mNegHalfWidth(-0.5f * width),
    mNegHalfHeight(-0.5f * height),
    mPosHalfWidth(0.5f * width),
    mPosHalfHeight(0.5f * height),
    mTargetWidth((F32)width),
    mTargetHeight((F32)height),
    mAngle(0),
    mScaleX(1), mScaleY(1),
    mPivotX(0), mPivotY(0)  // pivot reference point is center of image
{
    assert(graphics);
    if (graphics)
        mDriver = graphics->getIrrDevice()->getVideoDriver();
    mColor.set(0xffffffff);
    mUV[0].set(uv[0]);
    mUV[1].set(uv[1]);
    if (texture)
        texture->grab();
}


ImageIrr::~ImageIrr()
{
    if (mTexture)
        mTexture->drop();
}


void ImageIrr::setColor(const Color &color)
{
    mColor.set(color.a, color.r, color.g, color.b);
}


void ImageIrr::draw(IMAGE_DRAW_MODE drawMode, UI32 flags, F32 x, F32 y,
                    const Color &color)
{
    video::SColor irrcolor(color.a, color.r, color.g, color.b);
    draw(drawMode, flags, x, y, irrcolor);
}


void ImageIrr::draw(IMAGE_DRAW_MODE drawMode, UI32 flags, F32 x, F32 y,
                    const irr::video::SColor &color)
{
    // Note: IF_COLOR and IF_ALPHA flags are currently ignored in this implementation.

    video::SMaterial material = mGraphics->getDefaultMaterial();

    if (drawMode != IDM_NORMAL)
    {
        switch (drawMode)
        {
        case IDM_ADD:
            material.MaterialTypeParam = video::pack_texureBlendFunc(video::EBF_SRC_ALPHA, video::EBF_ONE, video::EMFN_MODULATE_1X);
            break;
        case IDM_MULTIPLY:
            material.MaterialTypeParam = video::pack_texureBlendFunc(video::EBF_ZERO, video::EBF_SRC_COLOR, video::EMFN_MODULATE_1X);
            break;
        }
    }
    if (flags & IF_WRAP_REPEAT)
    {
        // setFlag in API is just for bool in Irr 1.4, so we're setting the wrap manually
        //mDefaultMaterial.setFlag(video::EMF_TEXTURE_WRAP, video::ETC_REPEAT);
        //for (u32 a = 0; a < video::MATERIAL_MAX_TEXTURES; ++a)
        //    mDefaultMaterial.TextureLayer[a].TextureWrap = video::ETC_REPEAT;
        // enough to set just first texture layer as we're using EMT_ONETEXTURE_BLEND
        material.TextureLayer[0].TextureWrap = video::ETC_REPEAT;
    }
    if (flags & IF_FILTER_LINEAR)
        material.setFlag(video::EMF_BILINEAR_FILTER, true);

    material.setTexture(0, mTexture);
    mDriver->setMaterial(material);

    video::S3DVertex vertices[4];
    vertices[0].Color = color;
    vertices[1].Color = color;
    vertices[2].Color = color;
    vertices[3].Color = color;
    vertices[0].TCoords.set(mUV[0].X, mUV[0].Y);
    vertices[1].TCoords.set(mUV[1].X, mUV[0].Y);
    vertices[2].TCoords.set(mUV[0].X, mUV[1].Y);
    vertices[3].TCoords.set(mUV[1].X, mUV[1].Y);

    if (flags & IF_TRANSFORM)
    {
        core::matrix4 transform;

        transform.setTranslation(core::vector3df(x, y, 0));
        if (flags & IF_ROTATE)
            transform.setRotationRadians(core::vector3df(mAngle, 0, 0));
        if (flags & IF_SCALE)
            transform.setScale(core::vector3df(mScaleX, mScaleY, 0));

        if (flags & IF_CENTER)
        {
            vertices[0].Pos.set(mNegHalfWidth - mPivotX, mNegHalfHeight - mPivotY, 0);
            vertices[1].Pos.set(mPosHalfWidth - mPivotX, mNegHalfHeight - mPivotY, 0);
            vertices[2].Pos.set(mNegHalfWidth - mPivotX, mPosHalfHeight - mPivotY, 0);
            vertices[3].Pos.set(mPosHalfWidth - mPivotX, mPosHalfHeight - mPivotY, 0);
        }
        else
        {
            vertices[0].Pos.set(0, 0, 0);
            vertices[1].Pos.set((f32)mWidth, 0, 0);
            vertices[2].Pos.set(0, (f32)mHeight, 0);
            vertices[3].Pos.set((f32)mWidth, (f32)mHeight, 0);
        }
        transform.transformVect(vertices[0].Pos);
        transform.transformVect(vertices[1].Pos);
        transform.transformVect(vertices[2].Pos);
        transform.transformVect(vertices[3].Pos);
    }
    else
    {
        vertices[0].Pos.set(x, y, 0);
        vertices[1].Pos.set(x + mTargetWidth, y, 0);
        vertices[2].Pos.set(x, y + mTargetHeight, 0);
        vertices[3].Pos.set(x + mTargetWidth, y + mTargetHeight, 0);
    }

    u16 indices[6] = { 0, 1, 2,  3, 2, 1 };
    mDriver->drawVertexPrimitiveList(vertices, 4, indices, 2,
                                     video::EVT_STANDARD, scene::EPT_TRIANGLES);
}


} // namespace turska
#endif // T2_USE_IRR
